;**********************************************
; Code for SMBus communication, to be included in
; your program.
;
;
;**********************************************

	errorlevel  -302

#include "config.inc"


;**********************************************
; BEGIN - Validate settings
;**********************************************
#ifndef SMBUS_PORT_SDA
error	"SMBUS_PORT_SDA must be defined, for example: PORTA,0"
#endif
#ifndef SMBUS_PORT_SCL
error	"SMBUS_PORT_SCL must be defined, for example: PORTA,1"
#endif
#ifndef CLOCKSPEED
error	"CLOCKSPEED must be defined, for example: .8000000"
#endif
;**********************************************
; END - Validate settings
;**********************************************

; short cuts for frequently used commands
#define CLOCK_HIGH 		bsf SMBUS_PORT_SCL
#define CLOCK_LOW 		bcf SMBUS_PORT_SCL
#define DATA_HIGH		bsf SMBUS_PORT_SDA
#define DATA_LOW		bcf SMBUS_PORT_SDA
#define BIT				Bit, 0
#define ACK_BIT			Bit, 1


	udata_shr
TempSensorL res 1
TempSensorH res 1
TempObjectL res 1
TempObjectH res 1

	udata
Byte		res 1
Bit			res 1
Counter		res 1
DelayCounter1 res 1
DelayCounter2 res 1

	global	TempSensorL
	global	TempSensorH
	global	TempObjectL
	global	TempObjectH
	
SwitchToRead macro
	banksel	SMBUS_TRIS_BANK
	bsf		SMBUS_TRIS_SDA
	banksel	SMBUS_PORT_BANK
	endm
SwitchToWrite macro
	banksel	SMBUS_TRIS_BANK
	bcf		SMBUS_TRIS_SDA
	banksel	SMBUS_PORT_BANK
	DATA_LOW
	endm
	
	code
	
SMBus_Init
	global  SMBus_Init
	
	; give the sensor some time to initialize itself
	call	POR_Delay_25ms
	
	; set SDA to output
	SwitchToWrite

	; set SCL & SDA to high as the idle condition
	CLOCK_HIGH
	DATA_HIGH
	
	; maybe we don't need this, since the default is SMBus mode???
	; send "request condition" to disable PWM & switch to SMBus mode
	DATA_LOW
	call	SMBus_Delay_2ms ; should be at least 1.44ms
	DATA_HIGH
	call	SMBus_Delay_2ms ; includes the bus-free time of at least 4.7us
	
	; done config/init
	return

MLX90614_ReadSensorTemp
	global	MLX90614_ReadSensorTemp
	
	; start
	call	SMBusSendStart
	
	; send slave addr
	movlw	0x00		; b'0000000X' Slave addr + b'XXXXXXX0' ReadBit
	call	SMBusSendByte
	; send commando
	movlw	0x06		; Read sensor temp
	call	SMBusSendByte
	; send re-start
	call	SMBusSendReStart
	; send slave addr
	movlw	0x00		; b'0000000X' Slave addr + b'XXXXXXX0' ReadBit
	call	SMBusSendByte
	
	SwitchToRead
	; read LSB
	call	SMBus_ReadByte
	movwf	TempSensorL
	; read MSB
	call	SMBus_ReadByte
	movwf	TempSensorH
	; read PEC
	call	SMBus_ReadByte
	SwitchToWrite

	call	SMBus_Delay_7us
	call	SMBus_Delay_7us
	call	SMBus_Delay_7us
	call	SendStop
	return
	
MLX90614_ReadObjectTemp
	global	MLX90614_ReadObjectTemp
	; start
	call	SMBusSendStart
	
	; send slave addr
	movlw	0x00		; b'0000000X' Slave addr + b'XXXXXXX0' ReadBit
	call	SMBusSendByte
	; send commando
	movlw	0x07		; Read object temp
	call	SMBusSendByte
	; send re-start
	call	SMBusSendReStart
	; send slave addr
	movlw	0x00		; b'0000000X' Slave addr + b'XXXXXXX0' ReadBit
	call	SMBusSendByte
	
	SwitchToRead
	; read LSB
	call	SMBus_ReadByte
	movwf	TempObjectL
	; read MSB
	call	SMBus_ReadByte
	movwf	TempObjectH
	; read PEC
	call	SMBus_ReadByte
	SwitchToWrite

	call	SMBus_Delay_7us
	call	SMBus_Delay_7us
	call	SMBus_Delay_7us
	call	SendStop
	return

SMBusSendReStart
	global	SMBusSendReStart
	DATA_HIGH
	call	SMBus_Delay_7us
	CLOCK_HIGH
	call	SMBus_Delay_7us
SMBusSendStart
	global	SMBusSendStart
	; force SDA low while keeping SCL up
	DATA_LOW
	; wait T_hd:STA time (min 4.0us)
	call	SMBus_Delay_7us

	; set the clock low as well
	CLOCK_LOW
	; wait a bit more
	call	SMBus_Delay_7us
	call	SMBus_Delay_7us
	return
	
SendStop
	CLOCK_HIGH
	; wait a bit more
	call	SMBus_Delay_7us

	DATA_HIGH
	; wait a bit more
	call	SMBus_Delay_7us

	return

;*****
; Sends what is currently in W
; 
; The received ACK bit is set in ACK_BIT accordingly
;*****
SMBusSendByte
	global	SMBusSendByte
	
	movwf	Byte
	movlw	.8
	movwf	Counter
SMBusSendByte_Loop	
	bsf		BIT			; set the correct value into BIT
	btfss	Byte, 7		;  "
	bcf		BIT			;  "
	call	SendBit     ; write BIT to the bus
	rlf		Byte, F		; prepare for next round
	decfsz	Counter
	goto	SMBusSendByte_Loop
	
	; read the ACK bit
	SwitchToRead
	call	ReadBit
	bsf		ACK_BIT		; store the ACK bit into ACK_BIT
	btfss	BIT			;  "
	bcf		ACK_BIT		;  "
	SwitchToWrite		; switch back to normal state
	
	call	SMBus_Delay_7us ; wait a bit to show a cap between each byte+ack transaction
	return

;************
; Reads a byte from the bus and stores it into W
;
; Always send ACK (0)
;************
SMBus_ReadByte
	global	SMBus_ReadByte
	
	movlw	.8
	movwf	Counter
SMBus_ReadByte_Loop
	rlf		Byte, F
	call	ReadBit
	bsf		Byte, 0
	btfss	BIT
	bcf		Byte, 0
	decfsz	Counter
	goto	SMBus_ReadByte_Loop

	SwitchToWrite
	bcf		BIT			; ACK == 0
	call	SendBit
	SwitchToRead

	movfw	Byte
	return
		
SendBit
	; pre-condition is that the CLOCK is LOW
	
	; set the SDA line according to the requested value
	DATA_LOW
	btfsc	BIT
	DATA_HIGH

	call	SMBus_Delay_4_5us
	
	; next, wait T_su:dat (250ns) before raising the CLK
	if CLOCKSPEED == .4000000
		nop
	else 
		if CLOCKSPEED == .8000000
			nop
			nop
		else
			error	"Unimplemented clockspeed"
		endif
	endif
	
	; set clock to high, signalling that the value from SDA can now be read
	CLOCK_HIGH
	call	SMBus_Delay_7us ; HIGH_time should be between 4.0us and 50us
	
	; now the receiver is finished reading the SDA line, lower the CLK and get ready to send the next bit
	CLOCK_LOW
	call	SMBus_Delay_4_5us ; at least 4.7us, timeout after 27ms
	
	; done
	return

;*********
; Read one bit
; 
; Note: the caller should set SDA to input himself before calling this method
;
ReadBit
	; pre-condition is that the CLOCK is LOW
	call	SMBus_Delay_7us ; wait until the remote end has set the SDA accordingly

	; tell the other end that we are about to read the SDA line
	CLOCK_HIGH

	; read SDA line now
	bsf		BIT
	btfss	SMBUS_PORT_SDA
	bcf		BIT
	call	SMBus_Delay_7us ; wait a bit before pulling down the CLK
	
	CLOCK_LOW
	call	SMBus_Delay_4_5us ; wait a bit before continuing

	; done
	return
	
SMBus_Delay_4_5us
	if CLOCKSPEED == .4000000
		nop
	else
		if CLOCKSPEED == .8000000
			goto	$+1
			goto	$+1
			nop
		else
			error	"Unimplemented clockspeed"
		endif
	endif
	return

SMBus_Delay_7us
	if CLOCKSPEED == .4000000
		; 7 cycles
		goto	$+1
		goto	$+1
		goto	$+1
		nop
	else
		if CLOCKSPEED == .8000000
			; 14 cycles
			goto	$+1
			goto	$+1
			goto	$+1
			goto	$+1
			goto	$+1
			goto	$+1
			goto	$+1
		else
			error	"Unimplemented clockspeed"
		endif
	endif
	return

SMBus_Delay_2ms
	if CLOCKSPEED == .4000000
		;1998 cycles
		movlw	0x8F
		movwf	DelayCounter1
		movlw	0x02
		movwf	DelayCounter2
	else
		if CLOCKSPEED == .8000000
			;3998 cycles
			movlw	0x1F
			movwf	DelayCounter1
			movlw	0x04
			movwf	DelayCounter2
		else
			error	"Unimplemented clockspeed"
		endif
	endif
SMBus_Delay_2ms_loop
	decfsz	DelayCounter1, f
	goto	$+2
	decfsz	DelayCounter2, f
	goto	SMBus_Delay_2ms_loop
			;2 cycles
	goto	$+1
	return
	
POR_Delay_25ms
	if CLOCKSPEED == .4000000
		;24993 cycles
		movlw	0x86
		movwf	DelayCounter1
		movlw	0x14
		movwf	DelayCounter2
	else
		if CLOCKSPEED == .8000000
			;49993 cycles
			movlw	0x0E
			movwf	DelayCounter1
			movlw	0x28
			movwf	DelayCounter2
		else
			error	"Unimplemented clockspeed"
		endif
	endif
POR_Delay_25ms_loop
	decfsz	DelayCounter1, f
	goto	$+2
	decfsz	DelayCounter2, f
	goto	POR_Delay_25ms_loop
	;3 cycles
	goto	$+1
	nop
	;4 cycles (including call)
	return
	
	
	end